/* $Id: ifactor1.c,v 2.0.0.9 1998/06/09 11:47:19 belabas Exp belabas $ */
#include "pari.h"
static GEN N;

/*********************************************************************/
/**                                                                 **/
/**                        PSEUDO PRIMALITY                         **/
/**                                                                 **/
/*********************************************************************/
static GEN sqrt1, sqrt2, t1, t;
static long r1;

static void
init_miller(GEN n)
{
  N = (signe(n) < 0)? absi(n): n;
  t=subis(n,1); r1=vali(t); t1 = shifti(t,-r1);
  sqrt1=cgeti(lg(t)); sqrt1[1]=evalsigne(0)|evallgefint(2);
  sqrt2=cgeti(lg(t)); sqrt2[1]=evalsigne(0)|evallgefint(2);
}

/* is N strong pseudo-prime for base a ? `End matching' (check for square
 * roots of -1) added by GN */
static int
bad_for_base(GEN a)
{
  GEN c2, c = powmodulo(a,t1,N);
  long r;

  if (!is_pm1(c) && !egalii(t,c)) /* go fishing for -1, not for 1 */
  {
    for (r=r1-1; r; r--)	/* (this saves one squaring/reduction) */
    {
      c2=c; c=modii(sqri(c),N);
      if (egalii(t,c)) break;
    }
    if (!r) return 1;
    /* sqrt(-1) seen, compare or remember */
    if (signe(sqrt1))		/* we saw one earlier: compare */
    {
      /* check if too many sqrt(-1)s mod N */
      if (!egalii(c2,sqrt1) && !egalii(c2,sqrt2)) return 1; 
    }
    else { affii(c2,sqrt1); affii(subii(N,c2),sqrt2); } /* remember */
  }
  return 0;
}

/* Miller-Rabin test for k random bases */
long
millerrabin(GEN n, long k)
{
  long r,i,av2, av = avma;
  
  if (!signe(n)) return 0;
  /* If |n| <= 3, check if n = +- 1 */
  if (lgefint(n)==3 && (ulong)(n[2])<=3) return (n[2] != 1); 

  if (!mod2(n)) return 0;
  init_miller(n); av2=avma;
  for (i=1; i<=k; i++)
  {
    do r = smodsi(mymyrand(),n); while (!r);
    if (bad_for_base(stoi(r))) { avma=av; return 0; }
    avma=av2;
  }
  avma=av; return 1;
}

/* As above for k bases taken in pr (i.e not random).
 * We must have k<11 and |n|>2 (not checked).
 * If k > 4, any n < 25e9 is proven prime.
 */
static int
miller(GEN n, long k)
{
  long r,i,av2, av = avma;
  static long pr[] = {0, 2,3,5,7,11,13,17,19,23,29};

  if (!mod2(n)) return 0;
  init_miller(n); av2=avma;
  for (i=1; i<=k; i++)
  {
    r = smodsi(pr[i],n); if (!r) break;
    if (bad_for_base(stoi(r))) { avma = av; return 0; }
    avma=av2;
  }
  avma=av; return 1;
}

GEN
nextprime(GEN n)
{
  long toadd,av1,av2, av = avma;

  if (typ(n) != t_INT) n=gceil(n); /* accept arguments in R --GN */
  if (lgefint(n) <=3)
  { /* check if n<=3 */
    ulong k;
    if (!signe(n) || (k=n[2])<=2) { avma=av; return gdeux; }
    if (k == 3) return stoi(3);
  }
  /* here n>3 */
  switch(smodis(n,6))
  {
    case 0: n=addsi(1,n); toadd=4; break;
    case 1: toadd=4; break;
    case 2: n=addsi(3,n); toadd=2; break;
    case 3: n=addsi(2,n); toadd=2; break;
    case 4: n=addsi(1,n); toadd=2; break;
    case 5: toadd=2; break;
    default: return NULL; /* not reached */
  }
  av2 = av1 = avma;
  for(;;)
  {
    if (miller(n,10)) break;
    av1=avma; n=addsi(toadd,n); toadd = 6-toadd;
  }
  if (av1!=av2) return gerepile(av,av1,n);
  return (av1==av)? icopy(n): n;
}

GEN
precprime(GEN n)
{
  long tosub,av1,av2, av = avma;

  if (typ(n) != t_INT) n=gfloor(n); /* accept arguments in R --GN */
  if (lgefint(n) <= 3)
  { /* check if n<=4 */
    ulong k;
    if (!signe(n) || (k=n[2])<=1) { avma=av; return gzero; }
    if (k <= 3) { avma=av; return stoi(k); }
    if (k == 4) { avma=av; return stoi(3); }
  }
  /* here n>4 */
  switch(smodis(n,6))
  {
    case 0: n=addsi(-1,n); tosub=4; break;
    case 1: tosub=2; break;
    case 2: n=addsi(-1,n); tosub=2; break;
    case 3: n=addsi(-2,n); tosub=2; break;
    case 4: n=addsi(-3,n); tosub=2; break;
    case 5: tosub=4; break;
    default: return NULL; /* not reached */
  }
  av2 = av1 = avma;
  for(;;)
  {
    if (miller(n,10)) break;
    av1=avma; n=addsi(-tosub,n); tosub = 6-tosub;
  }
  if (av1!=av2) return gerepile(av,av1,n);
  return (av1==av)? icopy(n): n;
}

/***********************************************************************/
/**                                                                   **/
/**                        FACTORIZATION (ECM)                        **/
/**   Integer factorization using the elliptic curves method (ECM).   **/
/**   ellfacteur() returns a non trivial factor of N, assuming N>0,   **/
/**   is composite, and has no small prime divisor (P^-(N) > 65.536). **/
/**                                                                   **/
/***********************************************************************/
static GEN gl, *W;
#define nbc 10

static int
sisprime(long p)
{
  long av = avma, i = miller(stoi(p),5);
  avma = av; return i;
#if 0
  byteptr d = diffptr;
  long q = *d++;

  while (q*q <= p)
  {
    if (p%q == 0) return 0; 
    q += *d++;
  }
  return 1;
#endif
}

static int
elladd(GEN *X1, GEN *X2)
{
  GEN lambda, *Y1 = X1+nbc, *Y2 = X2+nbc, *A = W+nbc+1;
  long i, av=avma;
  for (i=0; i<nbc; i++)
  {
    subiiz(X1[i], X2[i], A[i]); if (signe(A[i])<0) addiiz(A[i],N,A[i]);
    modiiz(mulii(A[i], W[i]), N, W[i+1]);
  }
  avma=av; /* if gl != N we have a factor */
  if (!inversemodulo(W[nbc], N, &gl)) return !egalii(N,gl);
  while (i--)
  {
    lambda = modii(mulii(subii(Y1[i], Y2[i]), mulii(gl, W[i])), N);
    modiiz(subii(sqri(lambda), addii(X2[i], X1[i])), N, X2[i]);
    modiiz(subii(mulii(lambda, subii(X1[i], X2[i])), Y1[i]), N, Y2[i]);
    modiiz(mulii(gl, A[i]), N, gl);
  }
  avma=av; return 0;
}

static int
elldouble(GEN *X)
{
  GEN lambda,v, *Y = X+nbc;
  long i, av=avma;
  for (i=0; i<nbc; i++)
    modiiz(mulii(Y[i], W[i]), N, W[i+1]);
  avma = av; 
  if (!inversemodulo(W[nbc], N, &gl)) return !egalii(gl,N);
  while (i--)
  {
    lambda = modii(mulii(addsi(1, mulsi(3, mulii(X[i], X[i]))),
			 mulii(gl,W[i])), N);
    lambda = shifti(mpodd(lambda)? addii(lambda, N): lambda, -1);
    v = modii(subii(sqri(lambda), shifti(X[i],1)), N);
    modiiz(mulii(gl, Y[i]), N, gl);
    modiiz(subii(mulii(lambda, subii(X[i], v)), Y[i]), N, Y[i]);
    affii(v, X[i]);
  }
  avma = av; return 0;
}

static int
ellmult(long k, GEN *X, GEN *XAUX)
{
  if (k == 1)
  { 
    for (k = 2*nbc; k--; ) affii (X[k], XAUX[k]);
    return 0;
  }
  if (ellmult(k/2, X, XAUX) || elldouble(X)) return !egalii(gl,N);
  return k&1? elladd(XAUX,X): 0;
}

GEN
ellfacteur(GEN n, int insist)
{
  static long TB1[] =
    { 156,405,962,2240,4778,9004,18437,34155,66596,133297,
      247988,374990,649996,1170924,1967442,3276490,5249667 };
  long B1,B2,dsn,dsnmax,a,i,j,p,m,av, tf = lgefint(n);
  GEN *X,*XD,*XDD, res = cgeti(tf);

  av=avma; N = n; a = 1;
  W = 1 + (GEN*) cgeti(260*nbc + 2);
  X   = W  + 2*nbc + 1;
  XD  = X  + 2*nbc;
  XDD = XD + 2*nbc;
  {
    GEN w = cgeti(260*nbc*tf);
    W[0]=gun; m = evallg(tf) | evaltyp(t_INT);
    for (i = 260*nbc; i; i--) 
    {
      *w = m; W[i] = w; w += tf; /* hack for: w = cgeti(tf) */
    }
  }
  tf -= 2; dsn = (long)sqrt(9.6*tf); dsnmax = (long)(4.8*tf);
  if (dsnmax > 41) dsnmax = 41;
  if (dsn < 9) dsn = 9; else if (dsn > dsnmax) dsn = dsnmax;

  for(;;)
  {
    byteptr d = diffptr;

    for (i = 2*nbc; i--; ) affsi(a++, X[i]);
    B1 = TB1[dsn/2 - 4]; B2 = 50*B1;
    if (DEBUGLEVEL > 3)
    { 
      fprintferr("Time = %ld\tdsn = %d,\tB1 = %d,\tB2 = %d\n",
                 timer2(),dsn,B1,B2);
      flusherr(); 
    }
    for (p = *d++; p < B1; p += *d++)
    {
      for (m=B2; m>p; m/=p)
	if (ellmult(p,X,XD)) goto fin;
      if (! *d) break;
    }
    while (p < B1)
    {
      do p += 2; while (! sisprime(p));
      for (m=B2; m>p; m/=p)
	if (ellmult(p,X,XD)) goto fin;
    }
    for (i=0; i < 2*nbc; i++) affii (X[i], XD[i]);
    if (elldouble(XD) || ellmult(p,X,XDD)) goto fin;
    for (i=0; i < 2*nbc; i++) affii (XD[i], XDD[i]);
    if (elldouble(XDD)) goto fin;
    for (i=4; i < 256; i += 2)
    {
      for (j = (i-2)*nbc; j < i*nbc; j++) affii (XD[j], XDD[j]);
      if (elladd(XD, XD + i*nbc)) goto fin;
    }
    while (p < B2)
    {
      if (! *d) break;
      p += *d; if (elladd(X + nbc * *d++, X)) goto fin;
    }
    while (p < B2)
    {
      long oldp = p;
      do p += 2; while (! sisprime(p));
      if (elladd(X + (p-oldp)*nbc, X)) goto fin;
    }
    if (!insist) { avma = av; return NULL; }
    if (dsn < dsnmax) dsn += 2;
 }
 fin:
  affii(cmpii(sqri(gl),N) > 0 ? divii(N,gl): gl, res);
  if (DEBUGLEVEL > 3)
    { fprintferr("p = %d, facteur = %Z\n",p,(long)res); flusherr(); }
  avma=av; return res;
}
